/*
 * Copyright (C) 2020-2022  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/iface/report_user_service.hh"
#include "src/iface/tools/tools.hh"
#include "src/iface/tools/common_types_wrappers.hh"
#include "src/iface/tools/generic_wrappers.hh"
#include "src/iface/tools/generic_unwrappers.hh"
#include "src/iface/tools/run_report_unwrappers.hh"

#include "src/impl/exceptions.hh"
#include "src/impl/list_reports.hh"
#include "src/impl/run_report.hh"

#include "liblog/liblog.hh"

#include "libpg/pg_exception.hh"

#include <algorithm>
#include <cmath>

namespace Fred {
namespace DbReport {
namespace Iface {

namespace {

void append_rows(
        std::vector<Impl::Table::JsonEncodedFields>::const_iterator begin,
        std::vector<Impl::Table::JsonEncodedFields>::const_iterator end,
        Api::Table* table)
{
    auto& dst_rows = *(table->mutable_rows());
    dst_rows.Reserve(static_cast<int>(dst_rows.size() + (end - begin)));
    std::for_each(begin, end, [&](auto&& row)
    {
        Tools::wrap(row, dst_rows.Add());
    });
}

template <typename T>
class Crop
{
public:
    explicit constexpr Crop(T value) noexcept
        : value_{value}
    { }
    template <typename Tc>
    constexpr Tc at_most(Tc max) const noexcept
    {
        return value_ < static_cast<T>(max) ? static_cast<Tc>(value_)
                                            : max;
    }
private:
    T value_;
};

template <typename T>
constexpr auto crop(T value) noexcept
{
    return Crop<T>{value};
}

// sample of 256 rows to measure the average space consumption
void fill_sample(
        const Impl::Lib::RunReportReply& src,
        Api::Table* dst)
{
    const auto number_of_sample_rows = crop(src.table.rows.size()).at_most(256);
    const auto rows_begin = begin(src.table.rows);
    append_rows(rows_begin, rows_begin + number_of_sample_rows, dst);
}

auto get_number_of_rows_fitting_in_one_message(
    std::size_t number_of_rows,
    std::size_t space_used,
    std::size_t message_space_max)
{
    const auto result = (static_cast<double>(number_of_rows) * message_space_max) / space_used;
    return static_cast<int>(std::rint(result));
}

void fill_first_chunk(const Impl::Lib::RunReportReply& src, Api::Table* dst)
{
    dst->clear_header();
    dst->clear_rows();
    Tools::wrap(src.table.field_descriptions, dst->mutable_header()->mutable_field_descriptions());
    if (src.table.rows.empty())
    {
        return;
    }
    fill_sample(src, dst);
    static constexpr auto message_bytes_max = 4 * 1024 * 1024;
    // a bigger factor means smaller chunks and therefore more likely to fit in the message
    static constexpr auto reliability_factor = 4;
    const auto max_number_of_rows = get_number_of_rows_fitting_in_one_message(
            dst->rows_size(),
            dst->rows().SpaceUsedExcludingSelfLong(),
            message_bytes_max / reliability_factor);
    if (max_number_of_rows < 1)
    {
        throw std::runtime_error{"even just one line won't fit in the message"};
    }
    const bool chunk_too_big = max_number_of_rows < dst->rows_size();
    if (chunk_too_big)
    {
        // shrink the number of rows to reliably fit in a message limited by 4MB
        dst->mutable_rows()->DeleteSubrange(max_number_of_rows, dst->rows_size());
        return;
    }
    const auto requested_number_of_rows = crop(src.table.rows.size()).at_most(max_number_of_rows);
    const bool chunk_too_small = dst->rows_size() < requested_number_of_rows;
    if (chunk_too_small)
    {
        // expand the number of rows to reliably fit in a message limited by 4MB
        append_rows(
                begin(src.table.rows) + dst->rows_size(),
                begin(src.table.rows) + requested_number_of_rows,
                dst);
    }
};

class Sender
{
public:
    explicit Sender(const Impl::Table& src_table)
        : src_rows_begin_{begin(src_table.rows)},
          src_rows_end_{end(src_table.rows)}
    { }
    auto operator()(
            grpc::ServerWriter<Api::RunReportReply>* writer,
            const Api::RunReportReply& reply)
    {
        struct Finish
        {
            bool last_message_sended;
        };
        src_rows_begin_ += reply.data().table().rows_size();
        auto result = Finish{src_rows_begin_ == src_rows_end_};
        auto write_options = ::grpc::WriteOptions{};
        if (result.last_message_sended)
        {
            write_options.set_last_message();
        }
        writer->Write(reply, write_options);
        return result;
    }
    void fill_chunk(Api::Table* dst, int chunk_rows_max) const
    {
        dst->clear_header();
        dst->clear_rows();
        const auto rows_to_send = crop(src_rows_end_ - src_rows_begin_).at_most(chunk_rows_max);
        append_rows(src_rows_begin_, src_rows_begin_ + rows_to_send, dst);
    }
private:
    std::vector<Impl::Table::JsonEncodedFields>::const_iterator src_rows_begin_;
    std::vector<Impl::Table::JsonEncodedFields>::const_iterator src_rows_end_;
};

}//namespace Fred::DbReport::Iface::{anonymous}

grpc::Status ReportUserService::run_report(
        grpc::ServerContext*,
        const Api::RunReportRequest* api_request,
        grpc::ServerWriter<Api::RunReportReply>* writer)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "ReportUser", __func__);
    try
    {
        const auto result = [](const auto& api_request)
        {
            Impl::Lib::RunReportRequest impl_request;
            Tools::unwrap(api_request, impl_request);
            return Impl::run_report(impl_request);
        }(*api_request);

        Api::RunReportReply reply;
        fill_first_chunk(result, reply.mutable_data()->mutable_table());
        const auto chunk_rows_max = reply.mutable_data()->mutable_table()->rows_size();
        auto message_sender = Sender{result.table};
        while (!message_sender(writer, reply).last_message_sended)
        {
            message_sender.fill_chunk(reply.mutable_data()->mutable_table(), chunk_rows_max);
        }
        return grpc::Status::OK;
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_sql_error(e);
    }
    catch (const Impl::InvalidArgument& e)
    {
        return Tools::on_invalid_argument(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

grpc::Status ReportUserService::list_reports(
        grpc::ServerContext*,
        const google::protobuf::Empty*,
        Api::ListReportsReply* reply)
{
    LIBLOG_SET_CONTEXT(LogCtx, log_ctx, "ReportUser", __func__);
    try
    {
        const auto result = Impl::list_reports();
        Tools::wrap(result.data.reports, reply->mutable_data()->mutable_reports());
        return grpc::Status::OK;
    }
    catch (const Impl::OverloadedReportName& e)
    {
        return Tools::on_failed_precondition(e);
    }
    catch (const LibPg::ExecFailure& e)
    {
        return Tools::on_sql_error(e);
    }
    catch (const std::exception& e)
    {
        return Tools::on_std_exception(e);
    }
    catch (...)
    {
        return Tools::on_unknown_exception();
    }
}

}//namespace Fred::DbReport::Iface
}//namespace Fred::DbReport
}//namespace Fred
